package dev.fitko.fitconnect.core.submission;

import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionsForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.http.HttpHeaders;
import dev.fitko.fitconnect.core.http.MimeTypes;

import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

public class SubmissionApiService implements SubmissionService {

    private final OAuthService authService;
    private final HttpClient httpClient;
    private final ApplicationConfig config;

    public SubmissionApiService(final OAuthService authService, final HttpClient httpClient, final ApplicationConfig config) {
        this.authService = authService;
        this.httpClient = httpClient;
        this.config = config;
    }

    @Override
    public Destination getDestination(final UUID destinationID) throws RestApiException {

        final String url = String.format(config.getDestinationsEndpoint(), destinationID);

        try {
            return this.httpClient.get(url, getHeaders(), Destination.class).getBody();

        } catch (RestApiException e) {
            throw new RestApiException("Could not get destination " + destinationID, e);
        }
    }

    @Override
    public SubmissionForPickup announceSubmission(final CreateSubmission submission) throws RestApiException {

        try {
            return this.httpClient.post(config.getSubmissionsEndpoint(), getHeaders(), submission, SubmissionForPickup.class).getBody();
        } catch (RestApiException e) {
            throw new RestApiException("Could not announce submission for destination " + submission.getDestinationId(), e);
        }
    }

    @Override
    public void uploadAttachment(final UUID submissionId, final UUID attachmentId, final String encryptedAttachment) throws RestApiException {

        final String url = String.format(config.getAttachmentEndpoint(), submissionId, attachmentId);

        try {
            this.httpClient.put(url, getHeaders("application/jose"), encryptedAttachment, Void.class);
        } catch (RestApiException e) {
            throw new RestApiException("Could not upload attachment", e);
        }
    }

    @Override
    public String getAttachment(final UUID submissionId, final UUID attachmentId) throws RestApiException {

        final String url = String.format(config.getAttachmentEndpoint(), submissionId, attachmentId);

        try {
            return this.httpClient.get(url, getHeaders("application/jose"), String.class).getBody();
        } catch (RestApiException e) {
            throw new RestApiException("Could not get attachment " + attachmentId, e);
        }
    }

    @Override
    public Submission sendSubmission(final SubmitSubmission submission) throws RestApiException {

        final String url = String.format(config.getSubmissionEndpoint(), submission.getSubmissionId());

        try {
            return this.httpClient.put(url, getHeaders(), submission, Submission.class).getBody();
        } catch (RestApiException e) {
            throw new RestApiException("Could not send submission " + submission.getSubmissionId(), e);
        }
    }

    @Override
    public SubmissionsForPickup pollAvailableSubmissions(final int offset, final int limit) throws RestApiException {

        final String urlWithQueryParams = config.getSubmissionsEndpoint() + "?limit=" + limit + "&offset=" + offset;

        try {
            return this.httpClient.get(urlWithQueryParams, getHeaders(), SubmissionsForPickup.class).getBody();
        } catch (RestApiException e) {
            throw new RestApiException("Could not poll for available submissions", e);
        }
    }

    @Override
    public SubmissionsForPickup pollAvailableSubmissionsForDestination(final UUID destinationId, final int offset, final int limit) {

        final String urlWithQueryParams = config.getSubmissionsEndpoint() + "?destinationId=" + destinationId
                + "&limit=" + limit + "&offset=" + offset;

        try {
            return this.httpClient.get(urlWithQueryParams, getHeaders(), SubmissionsForPickup.class).getBody();
        } catch (RestApiException e) {
            throw new RestApiException("Could not poll for available submissions on destination " + destinationId, e);
        }
    }

    @Override
    public Submission getSubmission(final UUID submissionId) throws RestApiException {

        try {
            return this.httpClient.get(String.format(config.getSubmissionEndpoint(), submissionId), getHeaders(), Submission.class).getBody();
        } catch (RestApiException e) {
            throw new RestApiException("Submission with id " + submissionId + " does not exist", e, e.getStatusCode());
        }
    }

    private Map<String, String> getHeaders() {
        return getHeaders(MimeTypes.APPLICATION_JSON);
    }

    private Map<String, String> getHeaders(final String contentType) {

        return new HashMap<>(Map.of(
                HttpHeaders.AUTHORIZATION, "Bearer " + authService.getCurrentToken().getAccessToken(),
                HttpHeaders.CONTENT_TYPE, contentType,
                HttpHeaders.ACCEPT_CHARSET, StandardCharsets.UTF_8.toString()
        ));
    }
}
