package dev.fitko.fitconnect.core.keys;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.JWKSet;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.domain.auth.OAuthToken;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.RestEndpointBase;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.http.MimeTypes;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.text.ParseException;
import java.util.Map;
import java.util.UUID;

import static com.github.tomakehurst.wiremock.client.WireMock.containing;
import static com.github.tomakehurst.wiremock.client.WireMock.get;
import static com.github.tomakehurst.wiremock.client.WireMock.ok;
import static com.github.tomakehurst.wiremock.client.WireMock.urlEqualTo;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class PublicKeyServiceTest extends RestEndpointBase {

    private KeyService underTest;

    private OAuthService authServiceMock;

    private SubmissionService submissionServiceMock;
    private ValidationService validationServiceMock;

    @BeforeEach
    void setup() {
        authServiceMock = mock(OAuthService.class);
        submissionServiceMock = mock(SubmissionService.class);
        validationServiceMock = mock(ValidationService.class);

        final var baseUrl = "http://localhost:" + wireMockServer.port();

        final EnvironmentName envName = new EnvironmentName("TESTING");
        final Environment environment = new Environment();
        environment.setSubmissionBaseUrl(baseUrl);
        environment.setSelfServicePortalBaseUrl(baseUrl);
        environment.setAllowInsecurePublicKey(true);

        final ApplicationConfig config = new ApplicationConfig();
        config.setEnvironments(Map.of(envName, environment));
        config.setActiveEnvironment(envName);

        underTest = new PublicKeyService(config, new HttpClient(), authServiceMock, submissionServiceMock, validationServiceMock);
    }

    @Test
    void getPublicEncryptionKeyTest() throws IOException, ParseException {

        // Given
        final Destination destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setEncryptionKid("123");

        final OAuthToken authToken = new OAuthToken();
        authToken.setAccessToken("abc123");

        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");

        wireMockServer.stubFor(
                get(urlEqualTo("/v1/destinations/" + destination.getDestinationId() + "/keys/123"))
                        .withHeader("Authorization", containing("Bearer " + authToken.getAccessToken()))
                        .willReturn(ok()
                                .withBody(publicKeyJson)
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200)));

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);
        when(submissionServiceMock.getDestination(any())).thenReturn(destination);
        when(validationServiceMock.validatePublicKey(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final RSAKey rsaKey = underTest.getPublicEncryptionKey(destination.getDestinationId());

        // Then
        assertThat(RSAKey.parse(publicKeyJson), is(rsaKey));
    }

    @Test
    void getPublicEncryptionKeyWithInsecureKeyNotAllowedTest() throws IOException {

        // Given
        final Destination destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setEncryptionKid("123");

        final OAuthToken authToken = new OAuthToken();
        authToken.setAccessToken("abc123");

        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");

        final EnvironmentName envName = new EnvironmentName("PROD");
        final Environment environment = new Environment();
        environment.setSubmissionBaseUrl("http://localhost:" + wireMockServer.port());
        environment.setAllowInsecurePublicKey(false);

        final ApplicationConfig config = new ApplicationConfig();
        config.setEnvironments(Map.of(envName, environment));
        config.setActiveEnvironment(envName);

        final PublicKeyService keyService = new PublicKeyService(config, new HttpClient(), authServiceMock, submissionServiceMock, validationServiceMock);

        wireMockServer.stubFor(
                get(urlEqualTo("/v1/destinations/" + destination.getDestinationId() + "/keys/123"))
                        .withHeader("Authorization", containing("Bearer " + authToken.getAccessToken()))
                        .willReturn(ok()
                                .withBody(publicKeyJson)
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200)));

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);
        when(submissionServiceMock.getDestination(any())).thenReturn(destination);
        when(validationServiceMock.validatePublicKey(any(), any())).thenReturn(ValidationResult.error(new Exception("Public key is insecure !")));

        // When
        final InvalidKeyException exception = assertThrows(InvalidKeyException.class, () -> keyService.getPublicEncryptionKey(destination.getDestinationId()));

        // Then
        assertThat(exception.getMessage(), containsString("Invalid public encryption key"));
    }

    @Test
    void getPublicSignatureKeyTest() throws IOException, ParseException {

        // Given
        final Destination destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setEncryptionKid("123");

        final OAuthToken authToken = new OAuthToken();
        authToken.setAccessToken("abc123");

        final String expectedSignatureKey = getResourceAsString("/public_signature_test_key.json");

        wireMockServer.stubFor(
                get(urlEqualTo("/v1/destinations/" + destination.getDestinationId() + "/keys/123"))
                        .willReturn(ok()
                                .withBody(expectedSignatureKey)
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200)));

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);
        when(submissionServiceMock.getDestination(any())).thenReturn(destination);
        when(validationServiceMock.validatePublicKey(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final RSAKey retrievedSignatureKey = underTest.getPublicSignatureKey(destination.getDestinationId(), "123");

        // Then
        assertThat(retrievedSignatureKey, is(RSAKey.parse(expectedSignatureKey)));
    }

    @Test
    void getSubmissionServiceSignatureKeyTest() throws IOException, ParseException {

        // Given
        final Destination destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setEncryptionKid("123");

        final OAuthToken authToken = new OAuthToken();
        authToken.setAccessToken("abc123");

        final JWKSet jwkSet = new JWKSet(JWK.parse(getResourceAsString("/public_signature_test_key.json")));

        wireMockServer.stubFor(
                get(urlEqualTo("/.well-known/jwks.json"))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(jwkSet.toJSONObject()))
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200)));

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);
        when(validationServiceMock.validatePublicKey(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final RSAKey retrievedSignatureKey = underTest.getSubmissionServicePublicKey("123");

        // Then
        assertThat(retrievedSignatureKey, is(RSAKey.parse(jwkSet.getKeyByKeyId("123").toJSONObject())));
    }

    @Test
    void getPortalSignatureKeyTest() throws IOException, ParseException {

        // Given
        final Destination destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setEncryptionKid("123");

        final OAuthToken authToken = new OAuthToken();
        authToken.setAccessToken("abc123");

        final JWKSet jwkSet = new JWKSet(JWK.parse(getResourceAsString("/public_signature_test_key.json")));

        wireMockServer.stubFor(
                get(urlEqualTo("/.well-known/jwks.json"))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(jwkSet.toJSONObject()))
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200)));

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);
        when(validationServiceMock.validatePublicKey(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final RSAKey retrievedSignatureKey = underTest.getPortalPublicKey("123");

        // Then
        assertThat(retrievedSignatureKey, is(RSAKey.parse(jwkSet.getKeyByKeyId("123").toJSONObject())));
    }

    @Test
    void getSignatureKeyFromCustomUrlTest() throws IOException, ParseException {

        // Given
        final Destination destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setEncryptionKid("123");

        final OAuthToken authToken = new OAuthToken();
        authToken.setAccessToken("abc123");

        final JWKSet jwkSet = new JWKSet(JWK.parse(getResourceAsString("/public_signature_test_key.json")));

        wireMockServer.stubFor(
                get(urlEqualTo("/custom/path/.well-known/jwks.json"))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(jwkSet.toJSONObject()))
                                .withHeader("Content-Type", MimeTypes.APPLICATION_JSON)
                                .withStatus(200)));

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);
        when(validationServiceMock.validatePublicKey(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final RSAKey retrievedSignatureKey = underTest.getWellKnownKeysForSubmissionUrl("http://localhost:" + wireMockServer.port() + "/custom/path", "123");

        // Then
        assertThat(retrievedSignatureKey, is(RSAKey.parse(jwkSet.getKeyByKeyId("123").toJSONObject())));
    }


    private String getResourceAsString(final String filename) throws IOException {
        return new String(PublicKeyServiceTest.class.getResourceAsStream(filename).readAllBytes());
    }
}